from __future__ import annotations

from json import JSONEncoder
from typing import TYPE_CHECKING
from typing import Any

import orjson

from pysignalr.messages import CancelInvocationMessage  # 5
from pysignalr.messages import CloseMessage  # 7
from pysignalr.messages import CompletionMessage  # 3
from pysignalr.messages import HandshakeMessage
from pysignalr.messages import HandshakeRequestMessage
from pysignalr.messages import HandshakeResponseMessage
from pysignalr.messages import InvocationMessage  # 1
from pysignalr.messages import JSONMessage  # virtual
from pysignalr.messages import Message
from pysignalr.messages import MessageType
from pysignalr.messages import PingMessage  # 6
from pysignalr.messages import StreamInvocationMessage  # 4
from pysignalr.messages import StreamItemMessage  # 2
from pysignalr.protocol.abstract import Protocol

if TYPE_CHECKING:
    from collections.abc import Iterable


class MessageEncoder(JSONEncoder):
    """
    Custom JSONEncoder for encoding Message and MessageType objects.

    This class is a subclass of JSONEncoder and overrides the default() method
    to provide custom serialization for Message and MessageType objects.
    """

    def default(self, obj: Message | MessageType) -> str | int | dict[str, Any]:
        """
        Overrides the default() method for custom serialization.

        Args:
            obj (Message | MessageType): The object to be serialized.

        Returns:
            str | int | dict[str, Any]: The serialized object.
        """
        if isinstance(obj, MessageType):
            return obj.value
        return obj.dump()


message_encoder = MessageEncoder()


class BaseJSONProtocol(Protocol):
    """
    Base class for JSON protocols.

    This class provides the basic structure for JSON protocols and defines
    some abstract methods that must be implemented by subclasses.
    """

    def __init__(self) -> None:
        pass

    def decode(self, raw_message: str | bytes) -> tuple[JSONMessage]:
        """
        Decodes a raw message into a JSONMessage object.

        Args:
            raw_message (str | bytes): The raw message to be decoded.

        Returns:
            tuple[JSONMessage]: A tuple containing a single JSONMessage object.
        """
        json_message = orjson.loads(raw_message)
        return (JSONMessage(data=json_message),)

    def encode(self, message: Message | HandshakeRequestMessage) -> str | bytes:
        """
        Encodes a message into a raw representation.

        Args:
            message (Message | HandshakeRequestMessage): The message to be encoded.

        Returns:
            str | bytes: The raw representation of the message.
        """
        return orjson.dumps(message.dump())

    def decode_handshake(self, raw_message: str | bytes) -> tuple[HandshakeResponseMessage, Iterable[Message]]:
        """
        Decodes a handshake message.

        Args:
            raw_message (str | bytes): The raw handshake message to be decoded.

        Returns:
            tuple[HandshakeResponseMessage, Iterable[Message]]: A tuple containing a HandshakeResponseMessage and a sequence of Message objects.
        """
        raise NotImplementedError


class JSONProtocol(Protocol):
    """
    Class for handling JSON protocols.

    This class provides methods for encoding and decoding messages using the JSON protocol.
    """

    def __init__(self) -> None:
        super().__init__(
            protocol='json',
            version=1,
            record_separator=chr(0x1E),
        )

    def decode(self, raw_message: str | bytes) -> list[Message]:
        """
        Decodes a raw message into a list of Message objects.

        Args:
            raw_message (str | bytes): The raw message to be decoded.

        Returns:
            list[Message]: A list of Message objects.
        """
        if isinstance(raw_message, bytes):
            raw_message = raw_message.decode()

        raw_messages = raw_message.split(self.record_separator)
        messages: list[Message] = []

        for item in raw_messages:
            if item in ('', self.record_separator):
                continue

            dict_message = orjson.loads(item)
            if dict_message:
                messages.append(self.parse_message(dict_message))

        return messages

    def encode(self, message: Message | HandshakeMessage) -> str:
        """
        Encodes a message into a raw representation.

        Args:
            message (Message | HandshakeMessage): The message to be encoded.

        Returns:
            str: The raw representation of the message.
        """
        return message_encoder.encode(message) + self.record_separator

    def decode_handshake(self, raw_message: str | bytes) -> tuple[HandshakeResponseMessage, Iterable[Message]]:
        """
        Decodes a handshake message.

        Args:
            raw_message (str | bytes): The raw handshake message to be decoded.

        Returns:
            tuple[HandshakeResponseMessage, Iterable[Message]]: A tuple containing a HandshakeResponseMessage and a sequence of Message objects.
        """
        if isinstance(raw_message, bytes):
            raw_message = raw_message.decode()

        messages = raw_message.split(self.record_separator)
        messages = list(filter(bool, messages))
        data = orjson.loads(messages[0])
        idx = raw_message.index(self.record_separator)
        return (
            HandshakeResponseMessage(data.get('error', None)),
            self.decode(raw_message[idx + 1 :]) if len(messages) > 1 else [],
        )

    @staticmethod
    def parse_message(dict_message: dict[str, Any]) -> Message:
        """
        Parses a dictionary into a Message object.

        Args:
            dict_message (dict[str, Any]): The dictionary to be parsed.

        Returns:
            Message: The resulting Message object.
        """
        message_type = MessageType(dict_message.pop('type', 'close'))

        if message_type is MessageType.invocation:
            dict_message['invocation_id'] = dict_message.pop('invocationId', None)
            return InvocationMessage(**dict_message)
        elif message_type is MessageType.stream_item:
            return StreamItemMessage(**dict_message)
        elif message_type is MessageType.completion:
            dict_message['invocation_id'] = dict_message.pop('invocationId', None)
            return CompletionMessage(**dict_message)
        elif message_type is MessageType.stream_invocation:
            return StreamInvocationMessage(**dict_message)
        elif message_type is MessageType.cancel_invocation:
            return CancelInvocationMessage(**dict_message)
        elif message_type is MessageType.ping:
            return PingMessage()
        elif message_type is MessageType.close:
            dict_message['allow_reconnect'] = dict_message.pop('allowReconnect', None)
            return CloseMessage(**dict_message)
        else:
            raise NotImplementedError



from __future__ import annotations

# TODO: Refactor this module
from collections import deque
from typing import TYPE_CHECKING
from typing import Any
from typing import cast

import msgpack  # type: ignore[import-untyped]
import orjson

from pysignalr.messages import CancelInvocationMessage
from pysignalr.messages import CloseMessage
from pysignalr.messages import CompletionMessage
from pysignalr.messages import HandshakeRequestMessage
from pysignalr.messages import HandshakeResponseMessage
from pysignalr.messages import InvocationClientStreamMessage
from pysignalr.messages import InvocationMessage
from pysignalr.messages import Message
from pysignalr.messages import MessageType
from pysignalr.messages import PingMessage
from pysignalr.messages import StreamInvocationMessage
from pysignalr.messages import StreamItemMessage
from pysignalr.protocol.abstract import Protocol

if TYPE_CHECKING:
    from collections.abc import Iterable
    from collections.abc import Sequence

_attribute_priority = (
    # NOTE: Python limitation, left as is
    'type_',
    'type',
    'headers',
    'invocation_id',
    'target',
    'arguments',
    'item',
    'result_kind',
    'result',
    'stream_ids',
)


class MessagepackProtocol(Protocol):
    """
    Class for handling MessagePack protocols.

    This class provides methods for encoding and decoding messages using the MessagePack protocol.
    """

    def __init__(self) -> None:
        """
        Initializes a MessagepackProtocol object.
        """
        super().__init__(
            protocol='messagepack',
            version=1,
            record_separator=chr(0x1E),
        )

    def decode(self, raw_message: str | bytes) -> list[Message]:
        """
        Decodes a raw message into a list of Message objects.

        Args:
            raw_message (str | bytes): The raw message to be decoded.

        Returns:
            list[Message]: A list of Message objects.
        """
        messages: list[Message] = []
        offset = 0
        while offset < len(raw_message):
            length = msgpack.unpackb(raw_message[offset : offset + 1])
            values = msgpack.unpackb(raw_message[offset + 1 : offset + length + 1])
            offset += length + 1
            message = self.parse_message(values)
            messages.append(message)
        return messages

    def encode(self, message: Message | HandshakeRequestMessage) -> bytes:
        """
        Encodes a message into a raw representation.

        Args:
            message (Message | HandshakeRequestMessage): The message to be encoded.

        Returns:
            bytes: The raw representation of the message.
        """
        raw_message: deque[Any] = deque()

        for attr in _attribute_priority:
            if hasattr(message, attr):
                if attr == 'type_':
                    raw_message.append(getattr(message, attr).value)
                else:
                    raw_message.append(getattr(message, attr))

        encoded_message = cast('bytes', msgpack.packb(raw_message))
        varint_length = self._to_varint(len(encoded_message))
        return varint_length + encoded_message

    def decode_handshake(self, raw_message: str | bytes) -> tuple[HandshakeResponseMessage, Iterable[Message]]:
        """
        Decodes a handshake message.

        Args:
            raw_message (str | bytes): The raw handshake message to be decoded.

        Returns:
            tuple[HandshakeResponseMessage, Iterable[Message]]: A tuple containing a HandshakeResponseMessage and a sequence of Message objects.
        """
        if isinstance(raw_message, str):
            raw_message = raw_message.encode()

        has_various_messages = 0x1E in raw_message
        handshake_data = raw_message[0 : raw_message.index(0x1E)] if has_various_messages else raw_message
        messages = self.decode(raw_message[raw_message.index(0x1E) + 1 :]) if has_various_messages else []
        data = orjson.loads(handshake_data)
        return HandshakeResponseMessage(data.get('error', None)), messages

    @staticmethod
    def parse_message(seq_message: Sequence[Any]) -> Message:
        """
        Parses a sequence into a Message object.

        Args:
            seq_message (Sequence[Any]): The sequence to be parsed.

        Returns:
            Message: The resulting Message object.
        """
        # {} {'error'}
        # [1, Headers, InvocationId, Target, [Arguments], [StreamIds]]
        # [2, Headers, InvocationId, Item]
        # [3, Headers, InvocationId, ResultKind, Result]
        # [4, Headers, InvocationId, Target, [Arguments], [StreamIds]]
        # [5, Headers, InvocationId]
        # [6]
        # [7, Error, AllowReconnect?]

        msg = seq_message
        message_type = MessageType(msg[0])

        if message_type is MessageType.invocation:
            if len(msg[5]) > 0:
                return InvocationClientStreamMessage(headers=msg[1], stream_ids=msg[5], target=msg[3], arguments=msg[4])
            else:
                return InvocationMessage(headers=msg[1], invocation_id=msg[2], target=msg[3], arguments=msg[4])
        elif message_type is MessageType.stream_item:
            return StreamItemMessage(headers=msg[1], invocation_id=msg[2], item=msg[3])
        elif message_type is MessageType.completion:
            if msg[3] == 1:
                return CompletionMessage(headers=msg[1], invocation_id=msg[2], result=None, error=msg[4])
            elif msg[3] == 2:
                return CompletionMessage(headers=msg[1], invocation_id=msg[2], result=None, error=None)
            elif msg[3] == 3:
                return CompletionMessage(headers=msg[1], invocation_id=msg[2], result=msg[4], error=None)
            else:
                raise NotImplementedError
        elif message_type is MessageType.stream_invocation:
            return StreamInvocationMessage(headers=msg[1], invocation_id=msg[2], target=msg[3], arguments=msg[4])
        elif message_type is MessageType.cancel_invocation:
            return CancelInvocationMessage(headers=msg[1], invocation_id=msg[2])
        elif message_type is MessageType.ping:
            return PingMessage()
        elif message_type is MessageType.close:
            return CloseMessage(*msg[1:])
        else:
            raise NotImplementedError

    def _to_varint(self, value: int) -> bytes:
        """
        Converts an integer into a variable-length integer.

        Args:
            value (int): The integer to be converted.

        Returns:
            bytes: The variable-length integer.
        """
        buffer = b''

        while True:
            byte = value & 0x7F
            value >>= 7

            if value:
                buffer += bytes((byte | 0x80,))
            else:
                buffer += bytes((byte,))
                break

        return buffer



from __future__ import annotations

from dataclasses import dataclass
from enum import IntEnum
from typing import Any


@dataclass
class HandshakeMessage:
    """
    Base class for handshake messages.
    """

    def dump(self) -> dict[str, Any]:
        """
        Dumps the handshake message into a dictionary.

        Returns:
            dict[str, Any]: The dictionary representation of the handshake message.
        """
        return self.__dict__


@dataclass
class HandshakeRequestMessage(HandshakeMessage):
    """
    Handshake request message.

    Attributes:
        protocol (str): The protocol name.
        version (int): The protocol version.
    """

    protocol: str
    version: int


@dataclass
class HandshakeResponseMessage(HandshakeMessage):
    """
    Handshake response message.

    Attributes:
        error (str | None): The error message if the handshake failed.
    """

    error: str | None


class MessageType(IntEnum):
    """
    Enum representing the type of messages.

    Attributes:
        invocation (int): Invocation message type.
        stream_item (int): Stream item message type.
        completion (int): Completion message type.
        stream_invocation (int): Stream invocation message type.
        cancel_invocation (int): Cancel invocation message type.
        ping (int): Ping message type.
        close (int): Close message type.
        invocation_binding_failure (int): Invocation binding failure message type.
    """

    _ = 9999
    invocation = 1
    stream_item = 2
    completion = 3
    stream_invocation = 4
    cancel_invocation = 5
    ping = 6
    close = 7
    invocation_binding_failure = -1


@dataclass
class Message:
    """
    Base class for all messages.

    Methods:
        dump() -> dict[str, Any]: Dumps the message into a dictionary.
    """

    def __init_subclass__(cls, type_: MessageType) -> None:
        cls.type = type_  # type: ignore[attr-defined]

    def dump(self) -> dict[str, Any]:
        """
        Dumps the message into a dictionary.

        Returns:
            dict[str, Any]: The dictionary representation of the message.
        """
        data = self.__dict__

        invocation_id = data.pop('invocation_id', None)
        stream_ids = data.pop('stream_ids', None)

        data['type'] = self.type  # type: ignore[attr-defined]
        if invocation_id is not None:
            data['invocationId'] = invocation_id
        if stream_ids is not None:
            data['streamIds'] = stream_ids

        return data


@dataclass
class ResponseMessage(Message, type_=MessageType._):
    """
    Response message.

    Attributes:
        error (str | None): The error message.
        result (Any | None): The result of the message.
    """

    error: str | None
    result: Any | None


@dataclass
class CancelInvocationMessage(Message, type_=MessageType.cancel_invocation):
    """
    Cancel invocation message.

    Attributes:
        invocation_id (str): The ID of the invocation to cancel.
        headers (dict[str, Any] | None): Optional headers.
    """

    invocation_id: str
    headers: dict[str, Any] | None = None


@dataclass
class CloseMessage(Message, type_=MessageType.close):
    """
    Close message.

    Attributes:
        error (str | None): Optional error message.
        allow_reconnect (bool | None): Whether reconnection is allowed.
        headers (dict[str, Any] | None): Optional headers.
    """

    error: str | None = None
    allow_reconnect: bool | None = None
    headers: dict[str, Any] | None = None


@dataclass
class CompletionClientStreamMessage(Message, type_=MessageType.stream_item):
    """
    Completion client stream message.

    Attributes:
        invocation_id (str): The ID of the invocation.
        headers (dict[str, Any] | None): Optional headers.
    """

    invocation_id: str
    headers: dict[str, Any] | None = None


@dataclass
class CompletionMessage(Message, type_=MessageType.completion):
    """
    Completion message.

    Attributes:
        invocation_id (str): The ID of the invocation.
        result (Any | None): The result of the invocation.
        error (str | None): The error message if the invocation failed.
        headers (dict[str, Any] | None): Optional headers.
    """

    invocation_id: str
    result: Any | None = None
    error: str | None = None
    headers: dict[str, Any] | None = None

    def dump(self) -> dict[str, Any]:
        data = super().dump()

        result = data.pop('result', None)
        error = data.pop('error', None)
        headers = data.pop('headers', None)

        if result is not None:
            data['result'] = result
        if error is not None:
            data['error'] = error
        if headers is not None:
            data['headers'] = headers

        return data


@dataclass
class InvocationMessage(Message, type_=MessageType.invocation):
    """
    Invocation message.

    Attributes:
        invocation_id (str): The ID of the invocation.
        target (str): The target method name.
        arguments: The arguments for the method invocation.
        headers (dict[str, Any] | None): Optional headers.
    """

    invocation_id: str
    target: str
    arguments: Any
    headers: dict[str, Any] | None = None


@dataclass
class InvocationClientStreamMessage(Message, type_=MessageType.invocation):
    """
    Invocation client stream message.

    Attributes:
        stream_ids (list[str]): The stream IDs.
        target (str): The target method name.
        arguments: The arguments for the method invocation.
        headers (dict[str, Any] | None): Optional headers.
    """

    stream_ids: list[str]
    target: str
    arguments: Any
    headers: dict[str, Any] | None = None


@dataclass
class PingMessage(Message, type_=MessageType.ping):
    """
    Ping message.
    """

    pass


@dataclass
class StreamInvocationMessage(Message, type_=MessageType.stream_invocation):
    """
    Stream invocation message.

    Attributes:
        invocation_id (str): The ID of the invocation.
        target (str): The target method name.
        arguments: The arguments for the method invocation.
        headers (dict[str, Any] | None): Optional headers.
    """

    invocation_id: str
    target: str
    arguments: Any
    headers: dict[str, Any] | None = None


@dataclass
class StreamItemMessage(Message, type_=MessageType.stream_item):
    """
    Stream item message.

    Attributes:
        invocation_id (str): The ID of the invocation.
        item: The stream item.
        headers (dict[str, Any] | None): Optional headers.
    """

    invocation_id: str
    item: Any
    headers: dict[str, Any] | None = None


class JSONMessage(Message, type_=MessageType._):
    """
    JSON message used in BaseJSONProtocol to skip pysignalr-specific things.

    Attributes:
        data (dict[str, Any]): The JSON data.
    """

    def __init__(self, data: dict[str, Any]) -> None:
        self.data = data

    def dump(self) -> dict[str, Any]:
        """
        Dumps the JSON message into a dictionary.

        Returns:
            dict[str, Any]: The dictionary representation of the JSON message.
        """
        return self.data



from __future__ import annotations

from dataclasses import dataclass


@dataclass(frozen=True)
class HubError(Exception):
    """
    Base class for all Hub-related errors.
    """

    pass


@dataclass(frozen=True)
class AuthorizationError(HubError):
    """
    Exception raised for authorization errors.
    """

    pass


@dataclass(frozen=True)
class ConnectionError(HubError):
    """
    Exception raised for connection errors.

    Attributes:
        status (int): The HTTP status code related to the connection error.
    """

    status: int


@dataclass(frozen=True)
class ServerError(HubError):
    """
    Exception raised for server errors.

    Attributes:
        message (str | None): The error message from the server.
    """

    message: str | None


@dataclass(frozen=True)
class NegotiationFailure(HubError):
    """
    Exception raised when the protocol negotiation fails.
    """

    pass



from __future__ import annotations

from abc import ABC
from abc import abstractmethod
from typing import TYPE_CHECKING

from pysignalr.messages import HandshakeRequestMessage
from pysignalr.messages import HandshakeResponseMessage
from pysignalr.messages import Message

if TYPE_CHECKING:
    from collections.abc import Iterable


class Protocol(ABC):
    """
    Abstract base class for implementing a communication protocol.

    Attributes:
        protocol (str): Name of the protocol.
        version (int): Version of the protocol.
        record_separator (str): Record separator used in the protocol.
    """

    def __init__(self, protocol: str, version: int, record_separator: str) -> None:
        """
        Initializes a Protocol object.

        Args:
            protocol (str): Name of the protocol.
            version (int): Version of the protocol.
            record_separator (str): Record separator used in the protocol.
        """
        self.protocol = protocol
        self.version = version
        self.record_separator = record_separator

    @abstractmethod
    def decode(self, raw_message: str | bytes) -> Iterable[Message]:
        """
        Decodes a raw message into a sequence of messages.

        Args:
            raw_message (str | bytes): The raw message to be decoded.

        Returns:
            Iterable[Message]: A sequence of Message objects.
        """
        ...

    @abstractmethod
    def encode(self, message: Message | HandshakeRequestMessage) -> str | bytes:
        """
        Encodes a message into a raw representation.

        Args:
            message (Message | HandshakeRequestMessage): The message to be encoded.

        Returns:
            str | bytes: The raw representation of the message.
        """
        ...

    @abstractmethod
    def decode_handshake(self, raw_message: str | bytes) -> tuple[HandshakeResponseMessage, Iterable[Message]]:
        """
        Decodes a handshake message.

        Args:
            raw_message (str | bytes): The raw handshake message to be decoded.

        Returns:
            tuple[HandshakeResponseMessage, Iterable[Message]]: A tuple containing a HandshakeResponseMessage and a sequence of Message objects.
        """
        ...

    def handshake_message(self) -> HandshakeRequestMessage:
        """
        Creates a handshake message.

        Returns:
            HandshakeRequestMessage: A HandshakeRequestMessage object.
        """
        return HandshakeRequestMessage(self.protocol, self.version)



from abc import ABC
from abc import abstractmethod
from enum import IntEnum
from enum import auto

from pysignalr.messages import Message
from pysignalr.protocol.abstract import Protocol


class ConnectionState(IntEnum):
    """
    Enum representing the state of a connection.

    Attributes:
        connecting: The connection is being established.
        connected: The connection has been successfully established.
        reconnecting: The connection is being reestablished after being lost.
        disconnected: The connection has been lost or intentionally closed.
    """

    connecting = auto()
    connected = auto()
    reconnecting = auto()
    disconnected = auto()


class Transport(ABC):
    """
    Abstract base class for implementing a transport protocol.

    Attributes:
        protocol (Protocol): The protocol used by the transport.
        state (ConnectionState): The current state of the connection.
    """

    protocol: Protocol
    state: ConnectionState

    @abstractmethod
    async def run(self) -> None:
        """
        Abstract method for running the transport protocol.

        This method should be implemented by subclasses to handle the specifics of the transport protocol.
        """
        ...

    @abstractmethod
    async def send(self, message: Message) -> None:
        """
        Abstract method for sending a message.

        Args:
            message (Message): The message to be sent.

        This method should be implemented by subclasses to handle the specifics of the transport protocol.
        """
        ...



from __future__ import annotations

import asyncio
import logging
from http import HTTPStatus
from typing import TYPE_CHECKING

from aiohttp import ClientSession
from aiohttp import ClientTimeout
from aiohttp import ServerConnectionError
from websockets.asyncio.client import ClientConnection
from websockets.asyncio.client import connect
from websockets.exceptions import ConnectionClosed
from websockets.protocol import State

import pysignalr.exceptions as exceptions
from pysignalr.messages import CompletionMessage
from pysignalr.messages import Message
from pysignalr.messages import PingMessage
from pysignalr.transport.abstract import ConnectionState
from pysignalr.transport.abstract import Transport
from pysignalr.utils import get_connection_url
from pysignalr.utils import get_negotiate_url
from pysignalr.utils import replace_scheme

if TYPE_CHECKING:
    import ssl
    from collections.abc import Awaitable
    from collections.abc import Callable

    from pysignalr.protocol.abstract import Protocol

DEFAULT_MAX_SIZE = 2**20  # 1 MB
DEFAULT_PING_INTERVAL = 10
DEFAULT_CONNECTION_TIMEOUT = 10

DEFAULT_RETRY_SLEEP = 1
DEFAULT_RETRY_MULTIPLIER = 1.1
DEFAULT_RETRY_COUNT = 10

_logger = logging.getLogger('pysignalr.transport')


class WebsocketTransport(Transport):
    """
    WebsocketTransport is a class that manages WebSocket connections, handles sending and receiving messages,
    and manages connection states.

    Attributes:
        url (str): The URL of the WebSocket server.
        protocol (Protocol): The protocol used for message encoding/decoding.
        callback (Callable[[Message], Awaitable[None]]): The callback function to handle incoming messages.
        headers (dict[str, str] | None): Optional HTTP headers to include in the WebSocket handshake.
        skip_negotiation (bool): Whether to skip the negotiation step.
        ping_interval (int): The interval for sending ping messages to keep the connection alive.
        connection_timeout (int): The timeout for establishing a connection.
        max_size (int | None): The maximum size for incoming messages.
        access_token_factory (Callable[[], str] | None): A factory function to provide access tokens.
    """

    def __init__(
        self,
        url: str,
        protocol: Protocol,
        callback: Callable[[Message], Awaitable[None]],
        headers: dict[str, str] | None = None,
        skip_negotiation: bool = False,
        ping_interval: int = DEFAULT_PING_INTERVAL,
        connection_timeout: int = DEFAULT_CONNECTION_TIMEOUT,
        retry_sleep: float = DEFAULT_RETRY_SLEEP,
        retry_multiplier: float = DEFAULT_RETRY_MULTIPLIER,
        retry_count: int = DEFAULT_RETRY_COUNT,
        max_size: int | None = DEFAULT_MAX_SIZE,
        access_token_factory: Callable[[], str] | None = None,
        ssl: ssl.SSLContext | None = None,
    ):
        """
        Initializes the WebSocket transport with the provided parameters.

        Args:
            url (str): The URL of the WebSocket server.
            protocol (Protocol): The protocol used for message encoding/decoding.
            callback (Callable[[Message], Awaitable[None]]): The callback function to handle incoming messages.
            headers (dict[str, str] | None): Optional HTTP headers to include in the WebSocket handshake.
            skip_negotiation (bool): Whether to skip the negotiation step.
            ping_interval (int): The interval for sending ping messages to keep the connection alive.
            connection_timeout (int): The timeout for establishing a connection.
            max_size (int | None): The maximum size for incoming messages.
            access_token_factory (Callable[[], str] | None): A factory function to provide access tokens.
        """
        super().__init__()
        self._url = url
        self._protocol = protocol
        self._callback = callback
        self._headers = headers or {}
        self._skip_negotiation = skip_negotiation
        self._ping_interval = ping_interval
        self._connection_timeout = connection_timeout
        self._max_size = max_size
        self._access_token_factory = access_token_factory
        self._retry_sleep = retry_sleep
        self._retry_multiplier = retry_multiplier
        self._retry_count = retry_count
        self._ssl = ssl

        self._state = ConnectionState.disconnected
        self._connected = asyncio.Event()
        self._ws: ClientConnection | None = None
        self._open_callback: Callable[[], Awaitable[None]] | None = None
        self._close_callback: Callable[[], Awaitable[None]] | None = None

    def on_open(self, callback: Callable[[], Awaitable[None]]) -> None:
        """
        Registers a callback function to be called when the connection is opened.

        Args:
            callback (Callable[[], Awaitable[None]]): The callback function.
        """
        self._open_callback = callback

    def on_close(self, callback: Callable[[], Awaitable[None]]) -> None:
        """
        Registers a callback function to be called when the connection is closed.

        Args:
            callback (Callable[[], Awaitable[None]]): The callback function.
        """
        self._close_callback = callback

    def on_error(self, callback: Callable[[CompletionMessage], Awaitable[None]]) -> None:
        """
        Registers a callback function to be called when an error occurs.

        Args:
            callback (Callable[[CompletionMessage], Awaitable[None]]): The callback function.
        """
        self._error_callback = callback

    async def run(self) -> None:
        """
        Runs the WebSocket transport, managing the connection lifecycle.
        """
        while True:
            try:
                await self._loop()
            except exceptions.NegotiationFailure as e:
                await self._set_state(ConnectionState.disconnected)
                self._retry_count -= 1
                if self._retry_count <= 0:
                    raise e
                self._retry_sleep *= self._retry_multiplier
                await asyncio.sleep(self._retry_sleep)
            else:
                await self._set_state(ConnectionState.disconnected)

    async def send(self, message: Message) -> None:
        """
        Sends a message over the WebSocket connection.

        Args:
            message (Message): The message to be sent.
        """
        conn = await self._get_connection()
        await conn.send(self._protocol.encode(message))

    async def _loop(self) -> None:
        """
        Manages the connection lifecycle, including reconnection logic.
        """
        await self._set_state(ConnectionState.connecting)

        if not self._skip_negotiation:
            try:
                await self._negotiate()
            except ServerConnectionError as e:
                raise exceptions.NegotiationFailure from e

        # Since websockets interprets the presence of the ssl option as something different than providing None,
        # the call needs to be made with or without ssl option to work properly
        if self._ssl is None:
            connection_loop = connect(
                self._url,
                additional_headers=self._headers,
                ping_interval=self._ping_interval,
                open_timeout=self._connection_timeout,
                max_size=self._max_size,
                logger=_logger,
            )
        else:
            connection_loop = connect(
                self._url,
                additional_headers=self._headers,
                ping_interval=self._ping_interval,
                open_timeout=self._connection_timeout,
                max_size=self._max_size,
                logger=_logger,
                ssl=self._ssl,
            )

        async for conn in connection_loop:
            try:
                await self._handshake(conn)
                self._ws = conn
                await self._set_state(ConnectionState.connected)
                await asyncio.gather(
                    self._process(conn),
                    self._keepalive(conn),
                )

            except ConnectionClosed as e:
                _logger.warning('Connection closed: %s', e)
                self._ws = None
                await self._set_state(ConnectionState.reconnecting)

    async def _set_state(self, state: ConnectionState) -> None:
        """
        Sets the connection state and triggers appropriate callbacks.

        Args:
            state (ConnectionState): The new connection state.
        """
        if state == self._state:
            return

        _logger.info('State change: %s -> %s', self._state.name, state.name)

        if state == ConnectionState.connecting:
            if self._state != ConnectionState.disconnected:
                raise RuntimeError('Cannot connect while not disconnected')

            self._connected.clear()

        elif state == ConnectionState.connected:
            if self._state not in (ConnectionState.connecting, ConnectionState.reconnecting):
                raise RuntimeError('Cannot connect while not connecting or reconnecting')

            self._connected.set()

            if self._open_callback:
                await self._open_callback()

        elif state in (ConnectionState.reconnecting, ConnectionState.disconnected):
            self._connected.clear()

            if self._close_callback:
                await self._close_callback()

        else:
            raise NotImplementedError

        self._state = state

    async def _get_connection(self) -> ClientConnection:
        """
        Gets the active WebSocket connection, ensuring it is open.

        Returns:
            ClientConnection: The active WebSocket connection.

        Raises:
            RuntimeError: If the connection is closed or was never run.
        """
        try:
            await asyncio.wait_for(self._connected.wait(), self._connection_timeout)
        except TimeoutError as e:
            raise RuntimeError('The socket was never run') from e
        if not self._ws or self._ws.state != State.OPEN:
            raise RuntimeError('Connection is closed')
        return self._ws

    async def _process(self, conn: ClientConnection) -> None:
        """
        Processes incoming messages from the WebSocket connection.

        Args:
            conn (ClientConnection): The WebSocket connection.
        """
        while True:
            raw_message = await conn.recv()
            await self._on_raw_message(raw_message)

    async def _keepalive(self, conn: ClientConnection) -> None:
        """
        Sends periodic ping messages to keep the connection alive.

        Args:
            conn (ClientConnection): The WebSocket connection.
        """
        while True:
            await asyncio.sleep(10)
            await conn.send(self._protocol.encode(PingMessage()))

    async def _handshake(self, conn: ClientConnection) -> None:
        """
        Performs the WebSocket handshake with the server.

        Args:
            conn (ClientConnection): The WebSocket connection.
        """
        _logger.info('Sending handshake to server')
        token = self._access_token_factory() if self._access_token_factory else None
        if token:
            self._headers['Authorization'] = f'Bearer {token}'
        our_handshake = self._protocol.handshake_message()
        await conn.send(self._protocol.encode(our_handshake))

        _logger.info('Awaiting handshake from server')
        raw_message = await conn.recv()
        handshake, messages = self._protocol.decode_handshake(raw_message)
        if handshake.error:
            raise ValueError(f'Handshake error: {handshake.error}')
        for message in messages:
            await self._on_message(message)

    async def _negotiate(self) -> None:
        """
        Performs the negotiation step to establish the connection.
        """
        negotiate_url = get_negotiate_url(self._url)
        _logger.info('Performing negotiation, URL: `%s`', negotiate_url)

        session = ClientSession(
            timeout=ClientTimeout(connect=self._connection_timeout),
        )
        async with session:
            async with session.post(negotiate_url, headers=self._headers) as response:
                if response.status == HTTPStatus.OK:
                    data = await response.json()
                elif response.status == HTTPStatus.UNAUTHORIZED:
                    raise exceptions.AuthorizationError
                else:
                    raise exceptions.ConnectionError(response.status)

        connection_id = data.get('connectionId')
        url = data.get('url')
        access_token = data.get('accessToken')

        if connection_id:
            _logger.info('Negotiation completed')
            self._url = get_connection_url(self._url, connection_id)
        elif url and access_token:
            _logger.info('Negotiation completed (Azure)')
            self._url = replace_scheme(url, ws=True)
            self._headers['Authorization'] = f'Bearer {access_token}'
        else:
            raise exceptions.ServerError(str(data))

    async def _on_raw_message(self, raw_message: str | bytes) -> None:
        """
        Handles raw incoming messages, decoding them into protocol-specific messages.

        Args:
            raw_message (str | bytes): The raw incoming message.
        """
        for message in self._protocol.decode(raw_message):
            await self._on_message(message)

    async def _on_message(self, message: Message) -> None:
        """
        Handles decoded messages, passing them to the registered callback.

        Args:
            message (Message): The decoded message.
        """
        await self._callback(message)


class BaseWebsocketTransport(WebsocketTransport):
    """
    BaseWebsocketTransport is a subclass of WebsocketTransport that disables keepalive and handshake
    for simplified use cases.
    """

    async def _keepalive(self, conn: ClientConnection) -> None:
        """
        Disabled keepalive method.

        Args:
            conn (ClientConnection): The WebSocket connection.
        """
        return

    async def _handshake(self, conn: ClientConnection) -> None:
        """
        Disabled handshake method.

        Args:
            conn (ClientConnection): The WebSocket connection.
        """
        return



from __future__ import annotations

import urllib.parse as parse
from contextlib import suppress

http_schemas = ('http', 'https')
websocket_schemas = ('ws', 'wss')
http_to_ws = {k: v for k, v in zip(http_schemas, websocket_schemas, strict=False)}  # noqa: C416
ws_to_http = {k: v for k, v in zip(websocket_schemas, http_schemas, strict=False)}  # noqa: C416


def replace_scheme(url: str, ws: bool) -> str:
    """
    Replaces the scheme of a given URL from HTTP to WebSocket or vice versa.

    Args:
        url (str): The URL whose scheme is to be replaced.
        ws (bool): If True, replace HTTP with WebSocket schemes. If False, replace WebSocket with HTTP schemes.

    Returns:
        str: The URL with the replaced scheme.
    """
    scheme, netloc, path, query, fragment = parse.urlsplit(url)

    with suppress(KeyError):
        mapping = http_to_ws if ws else ws_to_http
        scheme = mapping[scheme]

    return parse.urlunsplit((scheme, netloc, path, query, fragment))


def get_negotiate_url(url: str) -> str:
    """
    Constructs the negotiation URL for the given SignalR endpoint URL.

    Args:
        url (str): The base SignalR endpoint URL.

    Returns:
        str: The negotiation URL.
    """
    scheme, netloc, path, query, fragment = parse.urlsplit(url)

    path = path.rstrip('/') + '/negotiate'
    with suppress(KeyError):
        scheme = ws_to_http[scheme]

    return parse.urlunsplit((scheme, netloc, path, query, fragment))


def get_connection_url(url: str, id: list[str]) -> str:
    """
    Constructs the connection URL with the given connection ID.

    Args:
        url (str): The base SignalR endpoint URL.
        id (list[str]): The connection ID.

    Returns:
        str: The connection URL.
    """
    scheme, netloc, path, query, fragment = parse.urlsplit(url)

    parsed_query = parse.parse_qs(query)
    parsed_query['id'] = id
    query = parse.urlencode(parsed_query, doseq=True)
    with suppress(KeyError):
        scheme = http_to_ws[scheme]

    return parse.urlunsplit((scheme, netloc, path, query, fragment))



from __future__ import annotations

import logging
import uuid
from collections import defaultdict
from collections.abc import AsyncIterator
from collections.abc import Awaitable
from collections.abc import Callable
from contextlib import asynccontextmanager
from typing import TYPE_CHECKING
from typing import Any

from pysignalr.exceptions import ServerError
from pysignalr.messages import CancelInvocationMessage
from pysignalr.messages import CloseMessage
from pysignalr.messages import CompletionClientStreamMessage
from pysignalr.messages import CompletionMessage
from pysignalr.messages import InvocationClientStreamMessage
from pysignalr.messages import InvocationMessage
from pysignalr.messages import Message
from pysignalr.messages import MessageType
from pysignalr.messages import PingMessage
from pysignalr.messages import StreamInvocationMessage
from pysignalr.messages import StreamItemMessage
from pysignalr.protocol.json import JSONProtocol
from pysignalr.transport.websocket import DEFAULT_CONNECTION_TIMEOUT
from pysignalr.transport.websocket import DEFAULT_MAX_SIZE
from pysignalr.transport.websocket import DEFAULT_PING_INTERVAL
from pysignalr.transport.websocket import DEFAULT_RETRY_COUNT
from pysignalr.transport.websocket import DEFAULT_RETRY_MULTIPLIER
from pysignalr.transport.websocket import DEFAULT_RETRY_SLEEP
from pysignalr.transport.websocket import WebsocketTransport

if TYPE_CHECKING:
    import ssl

    from pysignalr.protocol.abstract import Protocol
    from pysignalr.transport.abstract import Transport


EmptyCallback = Callable[[], Awaitable[None]]
AnyCallback = Callable[[Any], Awaitable[Any | None]]
MessageCallback = Callable[[Message], Awaitable[None | Any]]
CompletionMessageCallback = Callable[[CompletionMessage], Awaitable[None]]

_logger = logging.getLogger('pysignalr.client')


class ClientStream:
    """
    Client to server streaming implementation.
    Refer to https://docs.microsoft.com/en-gb/aspnet/core/signalr/streaming?view=aspnetcore-5.0#client-to-server-streaming
    for more information.

    Attributes:
        transport (Transport): The transport used to send stream items.
        target (str): The target method name on the server.
        invocation_id (str): The unique identifier for the stream.
    """

    def __init__(self, transport: Transport, target: str) -> None:
        self.transport: Transport = transport
        self.target: str = target
        self.invocation_id: str = str(uuid.uuid4())

    async def send(self, item: Any) -> None:
        """
        Sends the next item to the server.

        Args:
            item (Any): The item to send.
        """
        await self.transport.send(StreamItemMessage(self.invocation_id, item))

    async def invoke(self) -> None:
        """
        Starts the streaming process.
        """
        await self.transport.send(InvocationClientStreamMessage([self.invocation_id], self.target, []))

    async def complete(self) -> None:
        """
        Completes the streaming process.
        """
        await self.transport.send(CompletionClientStreamMessage(self.invocation_id))


class SignalRClient:
    """
    SignalRClient is a client for SignalR that manages connections, sends messages,
    and handles incoming messages.

    Attributes:
        url (str): The URL of the SignalR server.
        protocol (Protocol): The protocol used for message encoding/decoding.
        headers (dict[str, str]): Optional HTTP headers to include in the WebSocket handshake.
        access_token_factory (Callable[[], str] | None): A factory function to provide access tokens.
        _message_handlers (defaultdict[str, list[MessageCallback | None]]): Handlers for different message types.
        _stream_handlers (dict[str, tuple[MessageCallback | None, MessageCallback | None, CompletionMessageCallback | None]]): Handlers for stream messages.
        _invocation_handlers (dict[str, MessageCallback | None]): Handlers for invocation messages.
        _transport (WebsocketTransport): The transport used for WebSocket communication.
        _error_callback (CompletionMessageCallback | None): Callback for error messages.
    """

    def __init__(
        self,
        url: str,
        protocol: Protocol | None = None,
        headers: dict[str, str] | None = None,
        ping_interval: int = DEFAULT_PING_INTERVAL,
        connection_timeout: int = DEFAULT_CONNECTION_TIMEOUT,
        max_size: int | None = DEFAULT_MAX_SIZE,
        retry_sleep: float = DEFAULT_RETRY_SLEEP,
        retry_multiplier: float = DEFAULT_RETRY_MULTIPLIER,
        retry_count: int = DEFAULT_RETRY_COUNT,
        access_token_factory: Callable[[], str] | None = None,
        ssl: ssl.SSLContext | None = None,
    ) -> None:
        self._url = url
        self._protocol = protocol or JSONProtocol()
        self._headers = headers or {}
        self._access_token_factory = access_token_factory
        self._ssl = ssl

        self._message_handlers: defaultdict[str, list[MessageCallback | None]] = defaultdict(list)
        self._stream_handlers: dict[
            str, tuple[MessageCallback | None, MessageCallback | None, CompletionMessageCallback | None]
        ] = {}
        self._invocation_handlers: dict[str, MessageCallback | None] = {}

        self._transport = WebsocketTransport(
            url=self._url,
            protocol=self._protocol,
            callback=self._on_message,
            headers=self._headers,
            ping_interval=ping_interval,
            retry_sleep=retry_sleep,
            retry_multiplier=retry_multiplier,
            retry_count=retry_count,
            connection_timeout=connection_timeout,
            max_size=max_size,
            access_token_factory=access_token_factory,
            ssl=ssl,
        )
        self._error_callback: CompletionMessageCallback | None = None

    async def run(self) -> None:
        """
        Runs the SignalR client, managing the connection lifecycle.
        """
        await self._transport.run()

    def on(self, event: str, callback: AnyCallback) -> None:
        """
        Registers a callback function for a specific event.

        Args:
            event (str): The event name.
            callback (AnyCallback): The callback function.
        """
        self._message_handlers[event].append(callback)

    def on_open(self, callback: EmptyCallback) -> None:
        """
        Registers a callback function to be called when the connection is opened.

        Args:
            callback (EmptyCallback): The callback function.
        """
        self._transport.on_open(callback)

    def on_close(self, callback: EmptyCallback) -> None:
        """
        Registers a callback function to be called when the connection is closed.

        Args:
            callback (EmptyCallback): The callback function.
        """
        self._transport.on_close(callback)

    def on_error(self, callback: CompletionMessageCallback) -> None:
        """
        Registers a callback function to be called when an error occurs.

        Args:
            callback (CompletionMessageCallback): The callback function.
        """
        self._error_callback = callback

    async def send(
        self,
        method: str,
        arguments: list[dict[str, Any]],
        on_invocation: MessageCallback | None = None,
    ) -> None:
        """
        Sends a message to the server.

        Args:
            method (str): The method name to invoke on the server.
            arguments (list[dict[str, Any]]): The arguments to pass to the method.
            on_invocation (MessageCallback | None): Optional callback for the invocation response.
        """
        invocation_id = str(uuid.uuid4())
        message = InvocationMessage(invocation_id, method, arguments, self._headers)
        self._invocation_handlers[invocation_id] = on_invocation
        await self._transport.send(message)

    async def stream(
        self,
        event: str,
        event_params: list[str],
        on_next: MessageCallback | None = None,
        on_complete: MessageCallback | None = None,
        on_error: CompletionMessageCallback | None = None,
    ) -> None:
        """
        Starts a streaming invocation.

        Args:
            event (str): The event name to stream.
            event_params (list[str]): The parameters for the event.
            on_next (MessageCallback | None): Optional callback for each stream item.
            on_complete (MessageCallback | None): Optional callback when the stream is completed.
            on_error (CompletionMessageCallback | None): Optional callback for errors.
        """
        invocation_id = str(uuid.uuid4())
        message = StreamInvocationMessage(invocation_id, event, event_params, self._headers)
        self._stream_handlers[invocation_id] = (on_next, on_complete, on_error)
        await self._transport.send(message)

    @asynccontextmanager
    async def client_stream(self, target: str) -> AsyncIterator[ClientStream]:
        """
        Context manager for client-to-server streaming.

        Args:
            target (str): The target method name on the server.

        Yields:
            ClientStream: The client stream instance.
        """
        stream = ClientStream(self._transport, target)
        await stream.invoke()
        yield stream
        await stream.complete()

    async def _on_message(self, message: Message) -> None:
        """
        Handles incoming messages and routes them to the appropriate handlers.

        Args:
            message (Message): The incoming message.
        """
        if message.type == MessageType.invocation_binding_failure:  # type: ignore[attr-defined]
            raise ServerError(str(message))

        elif isinstance(message, PingMessage):
            pass

        elif isinstance(message, InvocationMessage):
            await self._on_invocation_message(message)

        elif isinstance(message, CloseMessage):
            await self._on_close_message(message)

        elif isinstance(message, CompletionMessage):
            await self._on_completion_message(message)

        elif isinstance(message, StreamItemMessage):
            await self._on_stream_item_message(message)

        elif isinstance(message, StreamInvocationMessage):
            pass

        elif isinstance(message, CancelInvocationMessage):
            await self._on_cancel_invocation_message(message)

        else:
            raise NotImplementedError

    async def _on_invocation_message(self, message: InvocationMessage) -> None:
        """
        Handles invocation messages.

        Args:
            message (InvocationMessage): The invocation message.
        """
        expects_response = message.invocation_id is not None
        callbacks = [callback for callback in self._message_handlers[message.target] if callback]

        if not callbacks:
            # There are no callbacks for the message.target
            _logger.warning("No client method with the name '%s' found.", message.target)
            if expects_response:
                _logger.error(
                    "No result given for '%s' method and invocation ID '%s'.",
                    message.target,
                    message.invocation_id,
                )
                await self._transport.send(
                    CompletionMessage(invocation_id=message.invocation_id, error="Client didn't provide a result.")
                )
            return None

        if expects_response and len(callbacks) > 1:
            # There are multiple callbacks, so multiple results for the message.target
            _logger.error("Multiple results provided for '%s'. Sending error to server.", message.target)
            await self._transport.send(
                CompletionMessage(invocation_id=message.invocation_id, error='Client provided multiple results.')
            )
            return None

        for callback in callbacks:
            try:
                res = await callback(message.arguments)
                if res:
                    if expects_response:
                        await self._transport.send(CompletionMessage(invocation_id=message.invocation_id, result=res))
                    else:
                        _logger.warning(
                            "Result given for '%s' method but server is not expecting a result.", message.target
                        )
                elif expects_response:
                    _logger.error(
                        "No result given for '%s' method and invocation ID '%s'.",
                        message.target,
                        message.invocation_id,
                    )
                    await self._transport.send(
                        CompletionMessage(
                            invocation_id=message.invocation_id,
                            error="Client didn't provide a result.",
                        )
                    )
            except Exception as exc:
                _logger.error("A callback for the method '%s' threw error '%s'.", message.target, exc)
                if not expects_response:
                    raise exc
                await self._transport.send(CompletionMessage(invocation_id=message.invocation_id, error=str(exc)))

        return None

    async def _on_completion_message(self, message: CompletionMessage) -> None:
        """
        Handles completion messages.

        Args:
            message (CompletionMessage): The completion message.
        """
        if message.error:
            if self._error_callback is None:
                raise RuntimeError('Error callback is not set')
            await self._error_callback(message)

        callback = self._invocation_handlers.pop(message.invocation_id)
        if callback is not None:
            await callback(message)

    async def _on_stream_item_message(self, message: StreamItemMessage) -> None:
        """
        Handles stream item messages.

        Args:
            message (StreamItemMessage): The stream item message.
        """
        callback, _, _ = self._stream_handlers[message.invocation_id]
        if callback:
            await callback(message.item)

    async def _on_cancel_invocation_message(self, message: CancelInvocationMessage) -> None:
        """
        Handles cancel invocation messages.

        Args:
            message (CancelInvocationMessage): The cancel invocation message.
        """
        _, _, callback = self._stream_handlers[message.invocation_id]
        if callback:
            await callback(message)  # type: ignore[arg-type]

    async def _on_close_message(self, message: CloseMessage) -> None:
        """
        Handles close messages.

        Args:
            message (CloseMessage): The close message.
        """
        if message.error:
            raise ServerError(message.error)



import asyncio
import importlib.metadata
import random
from collections.abc import AsyncIterator

import websockets.asyncio.client
from websockets.exceptions import InvalidHandshake

from pysignalr.exceptions import NegotiationFailure

__version__ = importlib.metadata.version('pysignalr')


# NOTE: As of `websockets` 12.0
BACKOFF_MIN = 1.92
BACKOFF_MAX = 60.0
BACKOFF_FACTOR = 1.618
BACKOFF_INITIAL = 5


async def __aiter__(
    self: websockets.asyncio.client.connect,
) -> AsyncIterator[websockets.asyncio.client.ClientConnection]:
    """
    Asynchronous iterator for the Connect object.

    This function attempts to establish a connection and yields the protocol when successful.
    If the connection fails, it retries with an exponential backoff.

    Args:
        self (websockets.asyncio.client.connect): The Connect object.

    Yields:
        websockets.asyncio.client.ClientConnection: The WebSocket protocol.

    Raises:
        NegotiationFailure: If the connection URL is no longer valid during negotiation.
    """
    backoff_delay = BACKOFF_MIN
    while True:
        try:
            async with self as protocol:
                yield protocol
        # Handle expired connection URLs by raising a NegotiationFailure exception.
        except (TimeoutError, InvalidHandshake) as e:
            raise NegotiationFailure from e

        except Exception:
            # Add a random initial delay between 0 and 5 seconds.
            # See 7.2.3. Recovering from Abnormal Closure in RFC 6544.
            if backoff_delay == BACKOFF_MIN:
                initial_delay = random.random() * BACKOFF_INITIAL
                self.logger.info(
                    '! connect failed; reconnecting in %.1f seconds',
                    initial_delay,
                    exc_info=True,
                )
                await asyncio.sleep(initial_delay)
            else:
                self.logger.info(
                    '! connect failed again; retrying in %d seconds',
                    int(backoff_delay),
                    exc_info=True,
                )
                await asyncio.sleep(int(backoff_delay))
            # Increase delay with truncated exponential backoff.
            backoff_delay = backoff_delay * BACKOFF_FACTOR
            backoff_delay = min(backoff_delay, BACKOFF_MAX)
            continue
        else:
            # Connection succeeded - reset backoff delay.
            backoff_delay = BACKOFF_MIN


# Override the __aiter__ method of the Connect class
websockets.asyncio.client.connect.__aiter__ = __aiter__  # type: ignore[method-assign]
